from iotbx import pdb
from cctbx.array_family import flex
from cctbx import sgtbx
from cctbx import uctbx
import scitbx.math
from scitbx import matrix
from libtbx.str_utils import show_string
from libtbx.test_utils import approx_equal
import sys, os

class conformer_info:

  def __init__(self, **keyword_args):
    self.__dict__.update(keyword_args)

class move_to_minimum_covering_sphere:

  def __init__(self, file_name, compute_start_radius_only=False):
    self.file_name = file_name
    self.pdb_input = pdb.input(file_name=file_name)
    print "source_info:", self.pdb_input.source_info()
    self.crystal_symmetry = self.pdb_input.crystal_symmetry_from_cryst1()
    assert self.crystal_symmetry is not None
    self.hierarchy = self.pdb_input.construct_hierarchy()
    self.conformer_infos = []
    assert self.hierarchy.models_size() == 1, "Not implemented."
    for model in self.hierarchy.models():
      for chain in model.chains():
        assert chain.conformers_size() == 1, "Not implemented."
        for conformer in chain.conformers():
          aa_sel = conformer.residue_class_selection(
            class_name="common_amino_acid")
          na_sel = conformer.residue_class_selection(
            class_name="common_rna_dna")
          sel = flex.union(
            size=len(conformer.residues()), iselections=[aa_sel, na_sel])
          f = conformer.select_residues(selection=sel.iselection())
          if (f.residues_size() > 1):
            residue_centers = f.residue_centers_of_geometry()
            minimum_covering_sphere = scitbx.math.minimum_covering_sphere(
              points=residue_centers, epsilon=0.1)
            print "chain="+show_string(chain.id), \
                  "conf="+show_string(f.id), \
                  residue_centers.size(), minimum_covering_sphere.radius()
            self.conformer_infos.append(conformer_info(
              chain_id=chain.id,
              conformer=f,
              residue_centers=residue_centers,
              minimum_covering_sphere=minimum_covering_sphere))
    self.keep_original = True
    if (len(self.conformer_infos) > 0):
      self.show_start_minimum_covering_sphere()
      if (not compute_start_radius_only):
        self.compute_asu_rt_matrices()
        if (1):
          self.loop_over_symmetry()
        else:
          self.try_all_permutations()
        if   (self.best_radius > self.start_radius*1.01):
          print "RESULT: PROBLEM: best_radius %.2f > %.2f start_radius" % (
            self.best_radius, self.start_radius)
        elif (self.best_radius < self.start_radius*0.99):
          print "RESULT: SUCCESS: best_radius %.2f < %.2f start_radius" % (
            self.best_radius, self.start_radius)
          self.keep_original = False
        else:
          print "RESULT: NO IMPROVEMENT"
        self.write_new_pdb()

  def show_start_minimum_covering_sphere(self):
    all_centers = flex.vec3_double()
    for cfi in self.conformer_infos:
      all_centers.extend(cfi.residue_centers)
    minimum_covering_sphere = scitbx.math.minimum_covering_sphere(
      points=all_centers, epsilon=0.1)
    self.start_radius = minimum_covering_sphere.radius()
    print "start radius:    %7.2f" % self.start_radius

  def compute_asu_rt_matrices(self):
    asu_mappings = self.crystal_symmetry.asu_mappings(buffer_thickness=0)
    sites_cart = flex.vec3_double([cfi.minimum_covering_sphere.center()
      for cfi in self.conformer_infos])
    asu_mappings.process_sites_cart(original_sites=sites_cart)
    fractionalize = self.crystal_symmetry.unit_cell().fractionalize
    for cfi,mapping in zip(self.conformer_infos,asu_mappings.mappings()):
      cfi.asu_rt_mx = asu_mappings.get_rt_mx(mapping[0])
      cfi.mcs_center_frac = fractionalize(cfi.minimum_covering_sphere.center())
    for i,cfi in enumerate(self.conformer_infos):
      cfi.index = i

  def loop_over_symmetry(self):
    uc = self.crystal_symmetry.unit_cell()
    orth = matrix.sqr(uc.orthogonalization_matrix())
    frac = matrix.sqr(uc.fractionalization_matrix())
    cfis = self.conformer_infos
    best_radius = None
    best_rtus = None
    for i_start in xrange(len(cfis)):
      print "start with:", i_start
      cfi = cfis[i_start]
      fixed = cfi.residue_centers.deep_copy()
      fixed_center_frac = cfi.mcs_center_frac
      rtus = [None]*len(cfis)
      rtus[i_start] = sgtbx.rt_mx().as_rational().as_float()
      moving = cfis[:i_start] + cfis[i_start+1:]
      smallest_radius = None
      while (len(moving) != 0):
        smallest_radius = None
        best_rctu = None
        best_rtu = None
        best_i_moving = None
        for i_moving,cfi in enumerate(moving):
          for s in self.crystal_symmetry.space_group():
            rt = s.add_unit_shifts_minimum_distance(
              site_frac_1=fixed_center_frac,
              site_frac_2=cfi.mcs_center_frac).as_rational().as_float()
            r = orth*rt.r*frac
            rc = r.elems * cfi.residue_centers
            for u0 in [-1,0,1]:
              for u1 in [-1,0,1]:
                for u2 in [-1,0,1]:
                  tu = orth*(rt.t+matrix.col((u0,u1,u2)))
                  rctu = rc + tu
                  fixed_size = fixed.size()
                  fixed.extend(rctu)
                  mcs = scitbx.math.minimum_covering_sphere(
                    points=fixed, epsilon=0.1)
                  fixed.resize(fixed_size)
                  radius = mcs.radius()
                  if (smallest_radius is None or smallest_radius > radius):
                    smallest_radius = radius
                    best_rctu = rctu
                    best_rtu = (r, tu)
                    best_i_moving = i_moving
                    best_center_frac = uc.fractionalize(mcs.center())
        print "smallest radius: %7.2f" % smallest_radius
        sys.stdout.flush()
        fixed.extend(best_rctu)
        fixed_center_frac = best_center_frac
        rtus[moving[best_i_moving].index] = matrix.rt(best_rtu)
        del moving[best_i_moving]
      if (smallest_radius is not None
            and (best_radius is None or smallest_radius < best_radius)):
        best_rtus = rtus
        best_radius = smallest_radius
    if (best_radius is None):
      self.best_radius = self.start_radius
      for cfi in cfis:
        cfi.rtu = sgtbx.rt_mx().as_rational().as_float()
    else:
      self.best_radius = best_radius
      for cfi,rtu in zip(cfis, best_rtus):
        cfi.rtu = rtu

  def try_all_permutations(self, u_range=1):
    u_range = range(-u_range, u_range+1)
    uc = self.crystal_symmetry.unit_cell()
    orth = matrix.sqr(uc.orthogonalization_matrix())
    frac = matrix.sqr(uc.fractionalization_matrix())
    cfis = self.conformer_infos
    best_radius = None
    best_rtus = None
    for perm in flex.permutation_generator(size=len(cfis)):
      i_start = perm[0]
      cfi = cfis[i_start]
      seed_radius = cfi.minimum_covering_sphere.radius()
      print "seed radius:     %7.2f" % seed_radius
      sa_rt = cfi.asu_rt_mx.as_rational().as_float()
      r = orth*sa_rt.r*frac
      tu = orth*sa_rt.t
      fixed = r.elems * cfi.residue_centers + tu.elems
      rtus = [None]*len(cfis)
      rtus[i_start] = (r, tu)
      for i_move in perm[1:]:
        cfi = cfis[i_move]
        smallest_radius = None
        best_rctu = None
        best_rtu = None
        for s in self.crystal_symmetry.space_group():
          s = s.mod_short()
          sa_rt = (s.multiply(cfi.asu_rt_mx)).as_rational().as_float()
          r = orth*sa_rt.r*frac
          rc = r.elems * cfi.residue_centers
          for u0 in u_range:
            for u1 in u_range:
              for u2 in u_range:
                tu = orth*(sa_rt.t+matrix.col((u0,u1,u2)))
                rctu = rc + tu
                fixed_size = fixed.size()
                fixed.extend(rctu)
                mcs = scitbx.math.minimum_covering_sphere(
                  points=fixed, epsilon=0.1)
                fixed.resize(fixed_size)
                radius = mcs.radius()
                if (smallest_radius is None or smallest_radius > radius):
                  smallest_radius = radius
                  best_rctu = rctu
                  best_rtu = (r, tu)
        print "current radius:  %7.2f" % smallest_radius
        fixed.extend(best_rctu)
        rtus[i_move] = best_rtu
        if (best_radius is not None and smallest_radius > best_radius):
          break
      if (best_radius is None or best_radius > smallest_radius):
        best_radius = smallest_radius
        print "best radius:     %7.2f" % best_radius
        best_rtus = rtus
      sys.stdout.flush()
    self.best_radius = best_radius
    for cfi,rtu in zip(cfis, best_rtus):
      cfi.rtu = matrix.rt(rtu)

  def write_new_pdb(self):
    if (self.keep_original):
      out_id = "stripped"
    else:
      out_id = "minimum_covering_sphere"
    self.output_file_name \
      = os.path.splitext(os.path.basename(self.file_name))[0] \
      + "_%s.pdb" % out_id
    print "output file:", self.output_file_name
    out = open(self.output_file_name, "w")
    print >> out, pdb.format_cryst1_record(
      crystal_symmetry=self.crystal_symmetry)
    serial = 0
    for cfi in self.conformer_infos:
      rt = cfi.rtu
      for residue in cfi.conformer.residues():
        for atom in residue.atoms():
          if (atom.hetero):
            record_name="HETATM"
          else:
            record_name="ATOM"
          serial += 1
          if (self.keep_original):
            site = atom.xyz
          else:
            site = (rt.r*matrix.col(atom.xyz)+rt.t).elems
          print >> out, pdb.format_atom_record(
            record_name=record_name,
            serial=serial,
            name=atom.name,
            altLoc=" ",
            resName=residue.name,
            chainID=cfi.chain_id,
            resSeq=residue.seq,
            iCode=residue.icode,
            site=site,
            occupancy=atom.occ,
            tempFactor=atom.b,
            segID=atom.segid,
            element=atom.element,
            charge=atom.charge)
      print >> out, "TER"
    print >> out, "END"
    out.close()

def run(args):
  for file_name in args:
    if (not os.path.isfile(file_name)): continue
    moved = move_to_minimum_covering_sphere(file_name=file_name)
    verify = move_to_minimum_covering_sphere(
      file_name=moved.output_file_name, compute_start_radius_only=True)
    if (moved.keep_original):
      assert approx_equal(moved.start_radius, verify.start_radius)
    else:
      assert approx_equal(moved.best_radius, verify.start_radius, 0.01)

if (__name__ == "__main__"):
  run(sys.argv[1:])
